module calculation_mod_omp
USE omp_lib
USE disc_mod
! Module containing functions useful in the
! calculation part of the program
implicit none

type :: Vector_c
	complex :: vector(3)
end type Vector_c

real, parameter :: pi=4.0*atan(1.0), c=299792458.0
complex, parameter :: im=(0.0, 1.0)

contains

function dot_prod_c(z1,z2)
! Function returning the dot product of two 
! complex vectors z1 and z2
complex , dimension(:) :: z1,z2
complex :: dot_prod_c
integer :: i

if (size(z1)/=size(z2))then
	print *, 'Error - not equal sized vectors'
	return
else
	dot_prod_c=(0.0, 0.0)
	do i=1, size(z1)
		dot_prod_c=dot_prod_c+z1(i)*z2(i)
	end do
	return
end if

end function

function dot_prod_r(r1,r2)
! Function returning the dot product of 
! two complex vectors z1 and z2
real , dimension(:) :: r1,r2
real :: dot_prod_r
integer :: i

if (size(r1)/=size(r2))then
	print *, 'Error - not equal sized vectors'
	return
else
	dot_prod_r=0.0
	do i=1, size(r1)
		dot_prod_r=dot_prod_r+r1(i)*r2(i)
	end do
	return
end if

end function

function G(p1, p2, i, j, k)
! Function returning the Greens tensor element i,j 
! when p1 is the observation point, p2 is the reference point 
! and k is the wavenumber

real, dimension(3) :: p1, p2, r_vec!, k_vec
real               :: k, R, factor
integer            :: i,j,m
complex            :: G, alpha, Z

do m=1,3
	r_vec(m)=p1(m)-p2(m)
end do
!k=sqrt(k_vec(1)**2+k_vec(2)**2+k_vec(3)**2)
R=sqrt( (p1(1)-p2(1))**2 + (p1(2)-p2(2))**2 + (p1(3)-p2(3))**2 )
alpha=( ( -k**2/(R**3) )-( 3*k*im/(R**4) )+( 3/(R**5) ) )

!print *, k*R/pi
!print *, cexp(im*dot_prod_r(k_vec, r_vec))
!print *, cexp(im*k*R)
!print *, k*R/dot_prod_r(k_vec, r_vec)

if (i .ne. j)then
	factor=((p1(i)-p2(i))*(p1(j)-p2(j))/((k**2)*4.0*pi))
	G=factor*exp(im*(k*R))*alpha
else	
	!new method
	Z=(R/k**2)*(((p1(i)-p2(i))**2)*alpha+(im*k-(1.0/R))/(R**2))
	G=exp(im*(k*R))*(1/(4*pi*R))*(1.0+Z)
end if
return
end function

function fn(r, struct, pair_n, area, ot)
! Function returning the function value of the n'th 
! RWG basis function at the point r
! struct is the discretization data structure, 
! pair_n is the pair of neighbouring elements 
! for the n'th basis function,
! area is the area of the element considered, 
! ot is 1 or 2 dependent on which
! element in the pair in which the point r is located.
! Works of course only for triangular elements
real            :: r(3), fn(3), p(3), area, L, &
                   p1(3), p2(3), L_vec(3)
type(Pair)      :: pair_n
type(Structure) :: struct
integer         :: i, j, ot, pos_neg

! Finds the corner (p+ or p- dependent on value of ot) 
! which is not common for the two elements in pair n
do i=1, 3
  if (struct%elements(pair_n%elements(ot))%corners(i) &
  /= pair_n%corners(1) &
  .and. struct%elements(pair_n%elements(ot))%corners(i) &
  /= pair_n%corners(2))then
    p(1:3)=struct%points(struct%elements(pair_n%elements(ot))&
    %corners(i))%point
  end if
end do 
p1(1:3)=struct%points(pair_n%corners(1))%point
p2(1:3)=struct%points(pair_n%corners(2))%point
L_vec(1:3)=p1-p2
L=sqrt(L_vec(1)**2+L_vec(2)**2+L_vec(3)**2)

! creating pos_neg from ot: (1 or 2) -> (1 or -1)
pos_neg=-1*(((ot-1)*ot)-1)

!print *, 'r:', r, 'p:', p, r-p
fn=pos_neg*(r-p)*(L/(2*area))
return
end function

function norm_vec_c(p1, p2, p3, mid)
! Function returning the complex normal vector 
! (zero imaginary part) of a discretization element
! pointing outwards from origo when p1,p2,p3 are 
! corners of the element.
real    :: p1(3), p2(3), p3(3), v1(3), v2(3), norm_vec_r(3), &
           l, mid(3) 
complex :: norm_vec_c(3)
integer :: i
logical :: correct_way

v2(1:3)=p3(1:3)-p1(1:3)
v1(1:3)=p2(1:3)-p1(1:3)
norm_vec_r=cross_r(v1,v2)
l=sqrt(norm_vec_r(1)**2+norm_vec_r(2)**2+norm_vec_r(3)**2)
norm_vec_c(1:3)=(/ cmplx(norm_vec_r(1)),cmplx(norm_vec_r(2)),&
cmplx(norm_vec_r(3)) /)
norm_vec_c=norm_vec_c/l

correct_way=.true.
do i=1,3
  if (mid(i)*norm_vec_r(i)<0)then
    if ((norm_vec_r(i)>0.1 .or. norm_vec_r(i)<-0.1) )then
      correct_way=.false.
      print *,'wrong way:', norm_vec_c, 'mid:', mid
    end if
  end if
end do
if (correct_way.eqv. .false.)then
		norm_vec_c=-1.0*norm_vec_c
end if

return
end function


function cross_c(v1,v2)
! Function returning the cross product of 
! two complex vectors v1 and v2
complex, dimension(3) :: v1, v2, cross_c

cross_c(1)=v1(2)*v2(3)-v1(3)*v2(2)
cross_c(2)=v1(3)*v2(1)-v1(1)*v2(3)
cross_c(3)=v1(1)*v2(2)-v1(2)*v2(1)

return
end function

function cross_r(v1,v2)
! Function returning the cross product of 
! two real vectors v1 and v2
real, dimension(3) :: v1, v2, cross_r

cross_r(1)=v1(2)*v2(3)-v1(3)*v2(2)
cross_r(2)=v1(3)*v2(1)-v1(1)*v2(3)
cross_r(3)=v1(1)*v2(2)-v1(2)*v2(1)

return
end function

function J_gen(k, k_hat, struct, area_list, my, E_in)
! Function calculating the surface current density, J, 
! by the Method of Weighted Residuals
! k is the wave number, k_hat is the direction of the wave vector
! struct is the discretization data structure,
! area_list contains the area of all the elements, 
! my is the permeability of region 1,
! E_inc contains the incoming E-field at r=/(0,0,0)/,
! E_inc=E_inc_hat*E_inc_0*exp(k dot r) 
integer                                :: t,l,m,n,s,q,dir,&
                                          n_el, n_pairs, info
integer, dimension(:), allocatable     :: ipiv
complex                                :: G_tensor(3,3), temp_c, &
                                          f_c1(3), f_c2(3), &
                                          f_c3(3), sum1(3), sum2, &
                                          sum3
real                                   :: k, f_r(3), my, E_in(3),&
                                          k_hat(3), temp_r(3), &
                                          least
real, dimension(:), allocatable        :: area_list
complex, dimension(:,:), allocatable   :: D_mat, D_mat_pre
real, dimension(:,:,:), allocatable    :: f_mat
complex, dimension(:), allocatable     :: b_vec, rs_vec
type(Vector_c) , dimension(:), allocatable :: J_gen, E_inc_list
type(Structure)                            :: struct

n_pairs=size(struct%neighbours)
n_el=size(struct%elements)
allocate(E_inc_list(n_el))
allocate(J_gen(n_el))
allocate(D_mat(n_pairs,n_pairs))
allocate(D_mat_pre(n_pairs,n_pairs))
allocate(f_mat(n_pairs,2,3))
allocate(b_vec(n_pairs))
allocate(ipiv(n_pairs))

allocate(rs_vec(n_pairs))

do m=1,n_pairs
  do t=1,2
    f_mat(m,t,1:3)=fn(struct%midpoints(struct%neighbours(m)% &
    elements(t))%point, struct, struct%neighbours(m), &
    area_list(struct%neighbours(m)%elements(t)), t)
  end do
end do

! Calculating the incoming E-field, E_inc, at all centroids
do m=1,n_el
  E_inc_list(m)%vector=E_in*exp(im*dot_prod_r(k*k_hat, &
  struct%midpoints(m)%point))
end do

do m=1,n_pairs  ! for all rows of the linear system
  if (modulo(m,100).eq.0) then 
    print *, 'pair nr', m, 'of total', n_pairs 
  end if
  do n=1,n_pairs  ! for all columns of the linear system
    sum2=(0.0,0.0)
    do t=1,2  ! for both elements of the outer integral
      sum1=(/ (0.0,0.0),(0.0,0.0),(0.0,0.0) /)
      do l=1,2  ! for both elements of the inner integral
        if (struct%neighbours(m)%elements(t) &
        .ne.struct%neighbours(n)%elements(l))then
          do dir=1,3 ! for all three spatial directions 
            do q=1,3
              G_tensor(dir,q)=G(struct%midpoints &
              (struct%neighbours(m)%elements(t))%point, &
              struct%midpoints(struct%neighbours(n)%elements(l))&
              %point, dir, q, k)
           end do
          end do
          f_c1(1:3)=cmplx(f_mat(n,l,1:3))
          do dir=1,3
            temp_c=dot_prod_c(f_c1,G_tensor(dir,1:3))
            sum1(dir)=sum1(dir)+area_list(struct%neighbours(n)&
            %elements(l))*temp_c
          end do 
        end if
      end do
      f_c2(1:3)=cmplx(f_mat(m,t,1:3))!cmplx(f_r(1:3))
      sum2=sum2+dot_prod_c(f_c2, sum1) &
      *area_list(struct%neighbours(m)%elements(t))
    end do		
    D_mat(m,n)=sum2
  end do
  sum3=(0.0,0.0)
  do t=1,2
    f_c3(1:3)=cmplx(f_mat(m,t,1:3))
    sum3=sum3+dot_prod_c(f_c3, E_inc_list(struct%neighbours(m)&
    %elements(t))%vector)*&
    area_list(struct%neighbours(m)%elements(t))
  end do
  b_vec(m)=sum3
  rs_vec(m)=sum3
end do
 
D_mat=D_mat*k*c*my/im
D_mat_pre=D_mat

D_mat(1,1:2)=(/ (2.0,0.0), (2.0,0.0)/)
D_mat(2,1:2)=(/ (3.0,0.0), (5.0,0.0)/)

D_mat_pre=D_mat

call CGESV(n_pairs, 1, D_mat, n_pairs, ipiv, b_vec, n_pairs, info)

temp_r(1)=0.0
temp_r(2)=0.0
do m=1, n_pairs
  temp_c=(0.0,0.0)
  do n=1, n_pairs
    temp_c=temp_c+D_mat_pre(m,n)*b_vec(n)
  end do
	!print *, temp_c, rs_vec(m), abs(temp_c-rs_vec(m))
	temp_r(1)=temp_r(1)+abs(temp_c-rs_vec(m))
	temp_r(2)=temp_r(2)+abs(rs_vec(m))
	!print *, abs(rs_vec(m))
end do

if (info .eq. 0)then
  print *, 'Linear system solved successfully!'
else
  print *, 'An error occured while solving the linear system'
  stop 
end if

print *, 'avg error:', temp_r(1)/temp_r(2)
!read(*,*)

! Calculating the resulting current densities 
! in all element centroids
do n=1, n_el
  J_gen(n)%vector=(/(0.0,0.0), (0.0,0.0), (0.0,0.0)/)
end do

do n=1,n_pairs
  do t=1,2
      J_gen(struct%neighbours(n)%elements(t))%vector= &
      J_gen(struct%neighbours(n)%elements(t))%vector &
      +b_vec(n)*f_mat(n,t,1:3)
  end do
end do
return
end function

function J_gen_quad(k, k_hat, struct, area_list, my, E_in)
! Function calculating the surface current density, J, 
! by the Method of Weighted Residuals using 
! the 3-point Gaussian quadrature formula
! The function is parallelized
! k is the wave number, k_hat is the direction of the wave vector
! struct is the discretization data structure,
! area_list contains the area of all the elements, 
! my is the permeability of region 1,
! E_inc contains the incoming E-field at r=/(0,0,0)/, 
! E_inc=E_inc_hat*E_inc_0*exp(k dot r) 
integer                                 :: t,l,m,n,s,p,q, &
                                           dir, n_el, n_pairs, &
                                           info, counter
integer, dimension(:), allocatable      :: ipiv
complex                                 :: G_tensor(3,3), temp_c,&
                                           f_c1(3), f_c2(3), & 
                                           f_c3(3), sum1(3), & 
                                           sum2, sum3, sum_q(3),&
                                           sum_p, sum_q2
real                                    :: k, f_r(3), my, E_in(3),&
                                           k_hat(3), temp_r(3),&
                                           least,r_ob(3), r_ref(3)
real, dimension(:), allocatable         :: area_list
complex, dimension(:,:), allocatable    :: D_mat, D_mat_pre
real, dimension(:,:,:), allocatable     :: f_mat2
real, dimension(:,:,:,:), allocatable   :: f_mat
complex, dimension(:), allocatable      :: b_vec, rs_vec
type(Vector_c) , dimension(:,:), allocatable  :: E_inc_list
type(Vector_c) , dimension(:), allocatable    :: J_gen_quad
type(Structure)                               :: struct

n_pairs=size(struct%neighbours)
n_el=size(struct%elements)
allocate(E_inc_list(n_el,3))
allocate(J_gen_quad(n_el*3))
allocate(D_mat(n_pairs,n_pairs))
allocate(D_mat_pre(n_pairs,n_pairs))
allocate(f_mat(n_pairs,2,3,3))
allocate(f_mat2(n_pairs,2,3))
allocate(b_vec(n_pairs))
allocate(ipiv(n_pairs))

allocate(rs_vec(n_pairs))

do m=1,n_pairs
  do t=1,2
    f_mat2(m,t,1:3)=fn(struct%midpoints(struct%neighbours(m)&
    %elements(t))%point, struct, struct%neighbours(m), &
    area_list(struct%neighbours(m)%elements(t)), t)
	  do q=1,3
      f_mat(m,t,q,1:3)=fn(struct%quadpoints(struct%neighbours(m)&
      %elements(t),q)%point, struct, struct%neighbours(m), & 
      area_list(struct%neighbours(m)%elements(t)), t)
    end do
  end do
end do

! Calculating the incoming E-field, E_inc, at all quadrature points
do m=1,n_el
  do q=1,3
    E_inc_list(m,q)%vector= &
    E_in*exp(im*dot_prod_r(k*k_hat,struct%quadpoints(m,q)%point))
  end do
end do

counter=0
!$omp parallel private ( m,n, t,p,l, q, dir, s, G_tensor) & 
!$omp private ( sum2, sum_p, sum1, sum_q, r_ob ) &
!$omp private ( r_ref, temp_c, f_c1, f_c2, f_c3, sum_q2, sum3) &
!$omp shared ( n_pairs, area_list, struct, b_vec, rs_vec, D_mat)
!$omp do
do m=1,n_pairs ! for all rows of the linear system
  counter=counter+1
  if (modulo(counter,100).eq.0) then 
    print *, 'pair nr', counter, 'of total', n_pairs 
  end if
  do n=1,n_pairs ! for all columns of the linear system
    sum2=(0.0,0.0)
    do t=1,2 ! for both elements of the outer integral
      sum_p=(0.0,0.0)
      do p=1,3
        sum1=(/ (0.0,0.0),(0.0,0.0),(0.0,0.0) /)
        do l=1,2 ! for both elements of the inner integral
          sum_q=(/ (0.0,0.0),(0.0,0.0),(0.0,0.0) /)
          do q=1,3
            r_ob=struct%quadpoints(struct%neighbours(m)&
            %elements(t),p)%point
            r_ref=struct%quadpoints(struct%neighbours(n)&
            %elements(l),q)%point
            if (vec_len(r_ob-r_ref)>1e-15)then
              do dir=1,3  ! for all three spatial directions 
                do s=1,3
                  G_tensor(dir,s)=G(r_ob, r_ref, dir, s, k)
                end do
              end do
              f_c1(1:3)=cmplx(f_mat(n,l,q,1:3))
              do dir=1,3
                temp_c=dot_prod_c(f_c1,G_tensor(dir,1:3))
                sum_q(dir)=sum_q(dir)+temp_c
              end do 
            end if
          end do !q
          sum1(1:3)=sum1(1:3)+ &
          area_list(struct%neighbours(n)%elements(l))*sum_q(1:3)
        end do !l
        f_c2(1:3)=cmplx(f_mat(m,t,p,1:3))
        sum_p=sum_p+dot_prod_c(sum1, f_c2)
      end do !p
      sum2=sum2+sum_p*(1.0/9.0)* &
      area_list(struct%neighbours(m)%elements(t))
    end do !t	
    D_mat(m,n)=sum2
  end do
  sum3=(0.0,0.0)
  do t=1,2
    sum_q2=(0.0,0.0)
    do q=1,3
      f_c3(1:3)=cmplx(f_mat(m,t,q,1:3))
      sum_q2=sum_q2+dot_prod_c(f_c3, &
      E_inc_list(struct%neighbours(m)%elements(t),q)%vector)
    end do
    sum3=sum3+sum_q2*(1.0/3.0)* &
    area_list(struct%neighbours(m)%elements(t))
  end do
  b_vec(m)=sum3
  rs_vec(m)=sum3
end do
!$omp end do
!$omp end parallel
 
D_mat=D_mat*k*c*my/im
D_mat_pre=D_mat

call CGESV(n_pairs, 1, D_mat, n_pairs, ipiv, b_vec, n_pairs, info)

temp_r(1)=0.0
temp_r(2)=0.0
do m=1, n_pairs
  temp_c=(0.0,0.0)
  do n=1, n_pairs
    temp_c=temp_c+D_mat_pre(m,n)*b_vec(n)
  end do
  temp_r(1)=temp_r(1)+abs(temp_c-rs_vec(m))
  temp_r(2)=temp_r(2)+abs(rs_vec(m))
end do

if (info .eq. 0)then
  print *, 'Linear system solved successfully!'
else
  print *, 'An error occured while solving the linear system'
  stop 
end if

print *, 'avg rel error:', temp_r(1)/temp_r(2)
!read(*,*)

! Calculating the resulting current densities in all element points
do n=1, n_el
  do q=1,3
    J_gen_quad((n-1)*3+q)%vector &
    =(/(0.0,0.0), (0.0,0.0), (0.0,0.0)/)
  end do
end do

print *, 'J_gen initialized to 0' 

do n=1,n_pairs
  do t=1,2
    do q=1,3
      J_gen_quad((struct%neighbours(n)%elements(t)-1)*3+q)%vector &
      =J_gen_quad((struct%neighbours(n)%elements(t)-1)*3+q)%vector &
      +b_vec(n)*f_mat(n,t,q,1:3)
    end do
  end do
end do

print *, 'J_gen_quad calculated'
return
end function

function vec_len(v)
  real		:: v(3), vec_len
  vec_len=sqrt(v(1)**2+v(2)**2+v(3)**2)
  return
end function

end module calculation_mod_omp
