module disc_mod
!USE calculation_mod !, only : vec_len
! Module defining the discretization data structure and 
! subroutines for filling the structure with data using 
! different discretization algorithms.
implicit none

type :: Point
  real :: point(3)
end type Point

type :: Element
  integer :: nr_of_corners
  integer, dimension(:), allocatable :: corners
end type Element

type :: Pair
  integer :: corners(2)
  integer :: elements(2)
end type Pair

type Structure
  type(Point), dimension(:), allocatable   :: points
  type(Element), dimension(:), allocatable :: elements
  type(Pair), dimension(:), allocatable    :: neighbours
  type(Point), dimension(:), allocatable   :: midpoints
  type(Point), dimension(:,:), allocatable :: quadpoints
end type structure

contains 

subroutine stripack_convert(struct, n)
! Generates a triangulation discretization using 
! stripack and converts it to fit into the struct variable.
type(Structure) :: struct
integer :: n, i

! variables for stripack discretization
integer ( kind = 4 ), parameter :: nrow = 9
real    ( kind = 8 ) ds(n)
integer ( kind = 4 ) ier
real    ( kind = 8 ) x(n)
real    ( kind = 8 ) y(n)
real    ( kind = 8 ) z(n)
real    ( kind = 8 ) sc
integer ( kind = 4 ) list(6*n)
integer ( kind = 4 ) lptr(6*n)
integer ( kind = 4 ) lend(n)
integer ( kind = 4 ) iwk(2*n)
integer ( kind = 4 ) lnew
real    ( kind = 8 ) rlat(n)
real    ( kind = 8 ) rlon(n)	
integer ( kind = 4 ) nt
integer ( kind = 4 ) ltri(nrow,2*n-4)

print *, 'Creating delaunay discretization with #points = ', n

call random_number ( harvest = rlat(1:n) )
call random_number ( harvest = rlon(1:n) )

rlat(1:n) = ( ( 1.0D+00 - rlat(1:n) ) * (  -90.0D+00 ) &
            +             rlat(1:n)   *     90.0D+00 )

rlon(1:n) = ( ( 1.0D+00 - rlon(1:n) ) * ( -180.0D+00 ) &
            +             rlon(1:n)   *    180.0D+00 )
sc = atan ( 1.0D+00 ) / 45.0D+00
x(1:n) = sc * rlon(1:n)
y(1:n) = sc * rlat(1:n)
call trans ( n, y, x, x, y, z )
call trmesh ( n, x, y, z, list, lptr, lend, lnew, iwk, iwk(n+1),&
              ds, ier )
!do i=1,n
! write (*,"(F10.5,F10.5,E15.5)"),x(i), y(i), z(i)
!end do
!call trprnt ( n, x, y, z, 0, nrow, nt, ltri)
call trlist ( n, list, lptr, lend, nrow, nt, ltri, ier )
!call trlprt ( n, x, y, z, 0, nrow, nt, ltri)
!do i=1,nt
! write (*,"(4i4)"), i, ltri(1, i), ltri(2,i), ltri(3,i)
!end do

!Allocating and defining struct
allocate(struct%points(n))
do i=1,n
  struct%points(i)%point=(/x(i), y(i), z(i)/)
end do
allocate(struct%elements(nt))
do i=1,nt
  struct%elements(i)%nr_of_corners=3
  allocate(struct%elements(i)%corners(struct%elements(i)&
  %nr_of_corners))
  struct%elements(i)%corners=(/ltri(1:struct%elements(i)&
  %nr_of_corners, i)/)
end do

call find_neighbours(struct)
call find_midpoints(struct)
call find_quadpoints(struct)

end subroutine

subroutine recursive_triangulation(struct, n)
! Creating symmetric triangular discretization recursively
! n=6,18,58,258,1026, ...
type(Structure) :: struct
integer :: n, i, ne, np_curr, ne_curr, lim, split_num 

print *, 'Creating discretization recursively with #points = ', n

allocate(struct%points(n))
struct%points(1)%point=(/ 1.0, 0.0, 0.0 /)
struct%points(2)%point=(/-1.0, 0.0, 0.0 /)
struct%points(3)%point=(/ 0.0, 1.0, 0.0 /)
struct%points(4)%point=(/ 0.0,-1.0, 0.0 /)
struct%points(5)%point=(/ 0.0, 0.0, 1.0 /)
struct%points(6)%point=(/ 0.0, 0.0,-1.0 /)

!ne=2*(n-2)
!split_num=1+int(log10((2+n)/8.0)/log10(4.0))
ne=2*(n-2) !int(8*4**(split_num-1))
!print *, split_num, ne
!read(*,*)

allocate(struct%elements(ne))
do i=1,ne
  struct%elements(i)%nr_of_corners=3
  allocate(struct%elements(i)%corners(3))		
end do
struct%elements(1)%corners=(/ 1, 5, 3/)
struct%elements(2)%corners=(/ 3, 5, 2/)
struct%elements(3)%corners=(/ 2, 5, 4/)
struct%elements(4)%corners=(/ 4, 5, 1/)
struct%elements(5)%corners=(/ 1, 3, 6/)
struct%elements(6)%corners=(/ 3, 2, 6/)
struct%elements(7)%corners=(/ 2, 4, 6/)
struct%elements(8)%corners=(/ 4, 1, 6/)

np_curr=6
ne_curr=8

lim=int(log10(ne/6.0)/log10(4.0))
do i=1, lim
  call recursive_split(struct, np_curr, ne_curr)
  ne_curr=ne_curr*4
  np_curr=(ne_curr/2)+2
end do

call find_neighbours(struct)
call find_midpoints(struct)
call find_quadpoints(struct)

end subroutine 

subroutine recursive_split(struct, n, ne)
integer                                   :: n, ne, i, counter, & 
                                             ind_a, ind_b, ind_c
type(Structure)                           :: struct
type(Element), dimension(:), allocatable  :: el_list
real                                      :: a(3),b(3),c(3), &
                                             v0(3), v1(3), v2(3)
logical                                   :: lo(3)

allocate(el_list(4*ne))
counter=n+1

do i=1,ne
lo=(/ .false., .false., .false. /)
a=normalize(0.5*(struct%points(struct%elements(i)%corners(1))&
%point+struct%points(struct%elements(i)%corners(3))%point ))
b=normalize(0.5*(struct%points(struct%elements(i)%corners(1))&
%point+struct%points(struct%elements(i)%corners(2))%point ))
c=normalize(0.5*(struct%points(struct%elements(i)%corners(2))&
%point+struct%points(struct%elements(i)%corners(3))%point ))
ind_a=point_in_list(a,struct%points)
lo(1)=ind_a.eq. -1
if (lo(1))then ! a is not present in list
  struct%points(counter)%point=a
  ind_a=counter
  counter=counter+1
end if

ind_b=point_in_list(b,struct%points)
lo(1)=ind_b.eq. -1
if (lo(1))then ! a is not present in list
  struct%points(counter)%point=b
  ind_b=counter
  counter=counter+1
end if

ind_c=point_in_list(c,struct%points)
lo(1)=ind_c.eq. -1
if (lo(1))then ! a is not present in list
  struct%points(counter)%point=c
  ind_c=counter
  counter=counter+1
end if
struct%elements(ne+(i-1)*3+1)%corners=&
(/ struct%elements(i)%corners(1), ind_b, ind_a/)	
struct%elements(ne+(i-1)*3+2)%corners=&
(/ ind_b,  struct%elements(i)%corners(2), ind_c/)	
struct%elements(ne+(i-1)*3+3)%corners=&
(/ ind_a,  ind_c, struct%elements(i)%corners(3)/)
struct%elements(i)%corners=(/ ind_a, ind_b, ind_c /)	
end do


end subroutine

function point_in_list(p, list)
! Function returning the index of the point p in list
! and returns -1 if it is not in the list
real																		:: p(3), temp(3)
type(Point), dimension(:), allocatable	:: list
integer																	:: i
integer																  :: point_in_list

point_in_list=-1
do i=1,size(list)
temp=list(i)%point-p
temp(1)=sqrt(temp(1)**2+temp(2)**2+temp(3)**2)
if (temp(1)<1e-30)then
  point_in_list=i
  return
end if
end do
return
end function

subroutine find_quadpoints(struct)
! Subroutine finding 3 points in each element
! for gauss quad integration and saving the
! resulting points in struct%quadpoints
! Compatible with only triangular elements
type(Structure) :: struct
integer         :: m, ne

ne=size(struct%elements)
allocate(struct%quadpoints(ne, 3))
do m=1,ne
  struct%quadpoints(m,1)%point=(2.0/3.0) & 
  *struct%points(struct%elements(m)%corners(1))%point &
  +(1.0/6.0)*struct%points(struct%elements(m)%corners(2))%point&
  +(1.0/6.0)*struct%points(struct%elements(m)%corners(3))%point
  struct%quadpoints(m,2)%point=(1.0/6.0) &
  *struct%points(struct%elements(m)%corners(1))%point &
  +(2.0/3.0)*struct%points(struct%elements(m)%corners(2))%point &
  +(1.0/6.0)*struct%points(struct%elements(m)%corners(3))%point
  struct%quadpoints(m,3)%point=(1.0/6.0) &
  *struct%points(struct%elements(m)%corners(1))%point &
  +(1.0/6.0)*struct%points(struct%elements(m)%corners(2))%point &
  +(2.0/3.0)*struct%points(struct%elements(m)%corners(3))%point 
end do
end subroutine

real function area(point_list)
! Function calculating the area of an element consisting of the 
! points contained by point_list.
! In present form only compatible with triangular elements, 
! but may be extended.
type(Point), dimension(:) :: point_list
real , dimension(3) :: v1, v2

if (size(point_list) .eq. 3)then
  v1=point_list(2)%point-point_list(1)%point
  v2=point_list(3)%point-point_list(1)%point
  area=0.5*sqrt((v1(2)*v2(3)-v1(3)*v2(2))**2 &
  +(v1(3)*v2(1)-v1(1)*v2(3))**2+(v1(1)*v2(2)-v1(2)*v2(1))**2)
  return
else if (1 .eq. 0) then
  print *, 'false!'
  area=-999
  return
endif
end function

subroutine find_midpoints(struct)
! Subroutine finding all centroids of the elements in struct 
! and saving the resulting points in struct%midpoints.
! Compatible with only triangular elements.
integer         :: ne, i, j, corners
type(Structure) :: struct
real            :: v1(3),v2(3),v3(3), mid(3)

ne=size(struct%elements)
if (allocated(struct%midpoints))then
  deallocate(struct%midpoints)
endif
allocate(struct%midpoints(ne))
do i=1,ne
  v1=struct%points(struct%elements(i)%corners(1))%point
  v2=struct%points(struct%elements(i)%corners(2))%point
  v3=struct%points(struct%elements(i)%corners(3))%point
  mid=(v1+v2+v3)/3.0
  struct%midpoints(i)%point=mid
end do
end subroutine

subroutine find_neighbours(struct)
! Subroutine finding all neighbours in struct and 
! saving the result in struct%neighbours.
! Compatible with arbitrary number of corners.
integer                               :: ne, i, j, d1,d2, nn, &
                                         out_arr(2)
type(Pair), dimension(:), allocatable :: temp_neighbours
type(Structure)                       :: struct

nn=0
ne=size(struct%elements)
if (allocated(struct%neighbours))then
  deallocate(struct%neighbours)
endif
allocate(struct%neighbours(3*ne/2))

allocate(temp_neighbours(3*ne/2))
do i=1,ne
  do j=i+1,ne
  d1=size(struct%elements(i)%corners)
  d2=size(struct%elements(j)%corners)
  out_arr=common_2(struct%elements(i)%corners, & 
  struct%elements(j)%corners, d1, d2 ) 
    if ( out_arr(2) /= -1 )then
      nn=nn+1
      struct%neighbours(nn)%elements=(/ i,j /)
      struct%neighbours(nn)%corners=(/out_arr(1:2)/)
    end if
  end do
end do

end subroutine

function common_2(arr1, arr2, dim1, dim2)
! function checking whether arr1 and arr2 have 2 common elements.
! If true, the function returns the two common elements
! If false, the function returns (-1, -1)
! Compatible with arbitrary number of dimensions
integer :: dim1, dim2, i, j, n_equal, common_2(2), n_different
integer :: arr1(dim1), arr2(dim2)
common_2=(/-1,-1/)

n_equal=0
do i=1,dim1
n_different=0
do j=1,dim2
  if (arr1(i) == arr2 (j))then
    n_equal=n_equal+1
    common_2(n_equal)=arr1(i)
  endif		
end do
end do
return

end function common_2


function normalize(v)
real :: v(3), normalize(3), l

l = sqrt(v(1)**2+v(2)**2+v(3)**2)
normalize(1:3)=v(1:3)/l 
return
end function normalize

subroutine area_dist(struct, disc_type, area_list)
! Subroutine generating the area distribution and
! printing the result to the file 'area_dist'.
real, parameter                 :: pi=4.0*atan(1.0)
integer, parameter              :: out_unit=20
integer                         :: n, ne, disc_type, i, ints, &
                                   indx
real                            :: exp_avg, mean, std_dev, &
                                   area_list(:), delta, start,&
                                   max_area
real, dimension(:), allocatable :: hist
type(Structure)                 :: struct

n=size(struct%points)
ne=size(struct%elements)

!exp_avg=sqrt(8*pi/ne)
exp_avg=4*pi/ne

mean=0.0
max_area=0.0
do i=1,ne
mean=mean+area_list(i)
end do
mean=mean/ne
std_dev=0.0
do i=1, ne
std_dev=std_dev+(area_list(i)-mean)**2
end do
std_dev=sqrt(std_dev/ne)

start=mean-3*std_dev
ints=int(ne/20)+1

allocate(hist(ints))
do i=1, ints
  hist(i)=0.0
end do
delta=6*std_dev/ints

do i=1, ne
  indx=int((area_list(i)-start)/delta)+1
  if (indx<=ints .and. indx>0)then
    hist(indx)=hist(indx)+1.0/ne
  end if
  if (area_list(i)>max_area)then
    max_area=area_list(i)
  end if
end do

print *, 'mean:', mean
print *, 'exp_avg:', exp_avg
print *, 'std_dev:', std_dev
print *, 'max_area',max_area
print *, 'delta', delta
print *, 'intervals:', ints

! Printing the resulting area-distribution to area_dist file
open(unit=out_unit,file="area_dist",action="write", &
status="replace")
write (out_unit, *) 'ne', ne
write (out_unit, *) 'disc_type', disc_type
write (out_unit, *) 'mean', mean
write (out_unit, *) 'std_dev', std_dev
write (out_unit, *) 'max_area',max_area

do i=1,ints
if (start+delta*i>=0.0)then
  !write (*, "(e10.2,e10.2)") start+delta*i, hist(i)
  write (out_unit, "(e14.5,e14.5)") start+delta*i, hist(i)
end if
end do

close (out_unit)
end subroutine

end module
